unit StepperMotor;

interface
uses forms,ExtCtrls,windows,classes,math,Buttons,sysutils,Series,mmtimer;

type



  TsmDirection = (smNone,smLeft,smRight);
  TsmStepEvent = procedure(Mindex:integer; direction: TsmDirection)of object;
  TsmCommandChange = procedure(Mindex:integer)of object;

  // program stuff
  psmProgramStep = ^TsmProgramStep;
  TsmProgramStep = record
    Steps: integer;
    Frequency: double;  // Hz
    RampTime: double; // seconds
    Direction: TsmDirection;
  end;
  TsmProgramStepArray = array[0..0]of TsmProgramStep;
  psmProgramSteps = ^TsmProgramStepArray;
  // --

  TsmProgram = class(TObject)
    capacity,count: integer;
    Items: psmProgramSteps;
    IsLooping: boolean;
    function Add:psmProgramStep;
    procedure Clear;
    function Insert(index:integer):psmProgramStep;
    procedure Delete(index:integer);
    constructor Create;
    destructor Destroy;override;
  end;

  TStepperMotor = class(TObject)
  private
    LastStepTime: double;
    RotateDir: TsmDirection;
    fIsExecuting: boolean;
    procedure TimerEvent(Sender: TObject);
    procedure SetIsExecuting(const Value: boolean);
  public
    NextStepTime,
    TimePaused,
    PreviousCmdTime,
    CurrentStepTime: double;
    WaitForFinish,
    HasFinished: boolean;



    LoopCount,
    FailedCnt,             // error occurred if > 0;
    Mindex: integer;
    State: byte;        //current state of each Motor (2 bits significant)
    OnStep: TsmStepEvent;  // control program feedback

    Timer: TMultiMediaTimer;//TRealTimer;
    StepProgram: TsmProgram;              // list of commands
    OnCmdChange: TsmCommandChange;        // command row advance
    OnProgramFinished: TsmCommandChange;

    ControlFrequency,
    MaxFrequency: double;
    CurrentCommand,
    CurrentStep: integer;
    IsRotating: boolean;
    StepLabel: TSpeedButton;
    StepCount: integer;

    ResetDirection: TsmDirection;
    function GetProgramStepTime: double;
    procedure IncCurrentStep;
    function GetDirection: TsmDirection;

    procedure ProgramStep;
    procedure Step(Direction: TsmDirection);
    procedure Rotate(Direction: TsmDirection);
    procedure StartProgram(Step: integer);
    procedure Stop; // stops program and rotate
    constructor Create(MotorIndex: integer); //0..3
    destructor Destroy;override;
    property IsExecuting: boolean read fIsExecuting write SetIsExecuting;
  end;


  function LeftStep(state: integer): integer;
  function RightStep(state: integer): integer;

implementation
uses msglog;



function LeftStep(state: integer): integer;
begin
  case state of
  0: result:=2;
  3: result:=1;
  2: result:=3;
  else result:=0;
  end;
end;

function RightStep(state: integer): integer;
begin
  case state of
  0: result:=1;
  1: result:=3;
  3: result:=2;
  else result:=0
  end;
end;

{ TStepperMotor }
constructor TStepperMotor.Create(MotorIndex: integer);
begin
  StepCount := 0;
  LastStepTime:=0;
  CurrentStepTime:=0;
  TimePaused:=0;               //code_me
  NextStepTime:=0;
  Mindex:= MotorIndex;

  Timer:= TMultiMediaTimer.Create;// TRealTimer.Create;
  Timer.Enabled:=false;
  Timer.Interval:=1;
  Timer.OnTimer:= TimerEvent;

  ResetDirection := smRight;
  StepProgram:= TsmProgram.Create;
  IsRotating:= false;
  IsExecuting:= false;
  HasFinished:=true;
  ControlFrequency:=10;
  MaxFrequency:=1000;
  CurrentCommand:=0;
  CurrentStep:=0;
  FailedCnt:=0;
end;

// for usb (scheduled port writes)
function TStepperMotor.GetProgramStepTime: double;
var a,t,s: double; n: integer; ActivateNextCommand: boolean;
begin
  if NextStepTime>0 then begin
    result := NextStepTime;
    exit;
  end;

  repeat
    ActivateNextCommand:=false;
    if CurrentCommand>=StepProgram.count then begin
      CurrentCommand:=0;
      CurrentStep:=0;
      inc(LoopCount);
      WaitForFinish := true;
      HasFinished := false;
      if not StepProgram.IsLooping or (StepProgram.count=0) then begin
        IsExecuting := false;
        if StepProgram.count=0 then HasFinished := true;
//        if Assigned(OnProgramFinished) then OnProgramFinished(Mindex);
        NextStepTime := 0;
        PreviousCmdTime := 0;
        result := 99999999999999;
        exit;
      end;
//      if Assigned(OnProgramFinished) then OnProgramFinished(Mindex);
    end;

    if CurrentStep>=StepProgram.Items[CurrentCommand].Steps then begin // advance command
      CurrentStep:=0;
      inc(CurrentCommand);
      ActivateNextCommand := true;
//      if assigned(OnCmdChange)then OnCmdChange(Mindex);
    end;
  until not ActivateNextCommand;

  with StepProgram.Items[CurrentCommand] do begin
    if RampTime>0 then begin
      a := Frequency/RampTime;        // a = v/t           acceleration
      s := 0.5*a*(RampTime*RampTime);
      n := floor(s);                // number of steps until max speed

      if 2*CurrentStep >= Steps then begin
        if CurrentStep >= Steps-n then begin
          t := sqrt(2*(Steps-CurrentStep)/a) - sqrt(2*(Steps-(CurrentStep+1))/a); // (s = 0.5at) calc time to next step
        end else
          t := 1/Frequency;
      end else
      if CurrentStep < n then begin
        t := sqrt(2*(CurrentStep+1)/a) - sqrt(2*(CurrentStep)/a);
      end else
        t := 1/Frequency;

      NextStepTime := t;
    end else begin
      if Frequency=0 then NextStepTime:= 0
      else NextStepTime := 1/Frequency;
    end;

    result := NextStepTime;
  end;
end;

procedure TStepperMotor.IncCurrentStep;
begin
  inc(CurrentStep);
  CurrentStepTime := CurrentStepTime + NextStepTime;
  NextStepTime:=0;
end;


function TStepperMotor.GetDirection: TsmDirection;
begin
  result := StepProgram.Items[CurrentCommand].Direction;
end;

// for lpt (direct port writes)
procedure TStepperMotor.ProgramStep;
var a,t,s: double; n: integer;
begin
  if CurrentCommand>=StepProgram.count then begin
    CurrentCommand:=0;
    CurrentStep:=0;
    if not StepProgram.IsLooping or (StepProgram.count=0) then begin
      IsExecuting := false;
      Timer.Enabled:=false;
      if Assigned(OnProgramFinished) then OnProgramFinished(Mindex);
      exit;
    end;
    if Assigned(OnProgramFinished) then OnProgramFinished(Mindex);
  end;

  if CurrentStep>=StepProgram.Items[CurrentCommand].Steps then begin // advance command
    CurrentStep:=0;
    inc(CurrentCommand);
    ProgramStep;
    if assigned(OnCmdChange)then OnCmdChange(Mindex);
    exit;
  end;

  with StepProgram.Items[CurrentCommand] do begin
    if RampTime>0 then begin
      a := Frequency/RampTime;        // a = v/t           acceleration
      s := 0.5*a*(RampTime*RampTime);
      n := floor(s);                // number of steps until max speed

      if 2*CurrentStep >= Steps then begin
        if CurrentStep >= Steps-n then begin
          t := sqrt(2*(Steps-CurrentStep)/a) - sqrt(2*(Steps-(CurrentStep+1))/a); // (s = 0.5at) calc time to next step
        end else
          t := 1/Frequency;
      end else
      if CurrentStep < n then begin
        t := sqrt(2*(CurrentStep+1)/a) - sqrt(2*(CurrentStep)/a);
      end else
        t := 1/Frequency;

      Timer.Interval := t;
    end else begin
      if Frequency=0 then Timer.Interval:= 0
      else Timer.Interval := 1/Frequency;
    end;
    Step(Direction);
  end;
  inc(CurrentStep);
end;

procedure TStepperMotor.TimerEvent(Sender: TObject);
begin
  if IsExecuting then begin
    ProgramStep;
  end else
  if IsRotating then begin
    Step(RotateDir);
  end;
end;

procedure TStepperMotor.Rotate(Direction: TsmDirection);
begin
  IsRotating := True;
  RotateDir:= Direction;
  if ControlFrequency<=0 then Timer.Interval := 10;
  Timer.Interval := 1/ControlFrequency;
  Timer.Enabled := true;
end;

procedure TStepperMotor.StartProgram(Step: integer);
begin
  IsExecuting := true;
  Timer.Interval:=0;
  Timer.Enabled:=true;
  ProgramStep;
end;

procedure TStepperMotor.Step(Direction: TsmDirection);
var t: double;
begin
  if MaxFrequency>0 then begin
    t:= Timer.GetTicks;
    if t<LastStepTime then begin
      Sleep(round(1000/MaxFrequency));
      inc(FailedCnt);
    end;
    LastStepTime := t + 1/MaxFrequency -0.001;
  end;

  if Direction = smRight then begin
    State:= RightStep(State);
    if Assigned(OnStep) then OnStep(Mindex,Direction);
    inc(StepCount);
    if StepLabel<>nil then StepLabel.Caption := IntToStr(StepCount);
  end else if Direction = smLeft then begin
    State:= LeftStep(State);
    if Assigned(OnStep) then OnStep(Mindex,Direction);
    dec(StepCount);
    if StepLabel<>nil then StepLabel.Caption := IntToStr(StepCount);
  end;
end;

procedure TStepperMotor.Stop;
begin
  CurrentStepTime:=0;
  PreviousCmdTime:=0;
  NextStepTime:=0;
  CurrentCommand:=0;
  CurrentStep:=0;
  IsExecuting:=false;
  IsRotating:=false;
  HasFinished:=true;
  Timer.Enabled:=false;
end;

destructor TStepperMotor.Destroy;
begin
  Timer.Free;
  StepProgram.Free;
  inherited;
end;


procedure TStepperMotor.SetIsExecuting(const Value: boolean);
begin
  if fIsExecuting = Value then exit;
  fIsExecuting := Value;
  if fIsExecuting then begin
    WaitForFinish := false;
    HasFinished := false;
    CurrentStepTime := 0.001*GetTickCount;
    LoopCount:=0;
  end;
end;

{ TsmProgram }
function TsmProgram.Add:psmProgramStep;
begin
  if count>=capacity then begin
    inc(capacity,64);
    ReallocMem(Items,capacity*sizeof(TsmProgramStep));
  end;
  result:=@Items[count];
  inc(count);
end;

procedure TsmProgram.Clear;
begin
  count:=0;
  capacity:=0;
  FreeMem(Items);
  Items:=nil;
end;

constructor TsmProgram.Create;
begin
  count:=0;
  capacity:=0;
  Items:=nil;
  IsLooping:= false;
end;

procedure TsmProgram.Delete(index:integer);
var i:integer;
begin
  if index<0 then exit;
  if index<count then begin
    dec(count);
    for i:=index to count-1 do begin
      CopyMemory(@Items[i],@Items[i+1],sizeof(TsmProgramStep));
    end;
  end;
end;

destructor TsmProgram.Destroy;
begin
  FreeMem(Items);
end;

function TsmProgram.Insert(index:integer):psmProgramStep;
var i:integer;
begin
  if (index>=count) or (index<0) then begin
    result:=Add;
    exit;
  end;

  if count>=capacity then begin
    inc(capacity,64);
    ReallocMem(Items,capacity*sizeof(TsmProgramStep));
  end;

  for i:=count-1 downto index do begin
    CopyMemory(@Items[i+1],@Items[i],sizeof(TsmProgramStep));
  end;

  result:=@Items[index];
  inc(count);
end;



end.
